import json

import pint

import opensemantic.characteristics.quantitative._model as q
from opensemantic.characteristics.quantitative._static import quantity_registry


def test_addition():
    q1 = q.Mobility(
        value=1.0,
        unit=q.MobilityUnit.meter_squared_per_second_per_volt,
    )

    q2 = q.Mobility(
        value=1.0,
        unit=q.MobilityUnit.centi_meter_squared_per_second_per_volt,
    )

    q3 = q1 + q2
    assert (
        q3.value == 1.0001
        and q3.unit == q.MobilityUnit.meter_squared_per_second_per_volt
    )


def test_quantity_value():
    q1 = q.Length(value=1.0, unit=q.LengthUnit.milli_meter)

    q_json = json.loads(q1.json(exclude_none=True))
    # print(q_json)
    assert q_json == {
        "type": ["Category:OSWee9c7e5c343e542cb5a8b4648315902f"],
        "value": 1.0,
        "unit": (
            "Item:OSWf101d25e944856e3bd4b4c9863db7de2"
            "#OSW322dec469be75aedb008b3ebff29db86"
        ),
    }

    q1 = q.Length(value=1.0, unit=q.LengthUnit.meter)

    q_json = json.loads(q1.json(exclude_none=True))
    # print(q_json)
    assert q_json == {
        "type": ["Category:OSWee9c7e5c343e542cb5a8b4648315902f"],
        "value": 1.0,
        "unit": "Item:OSWf101d25e944856e3bd4b4c9863db7de2",
    }

    q_json = json.loads(q1.json(exclude_none=True, exclude_defaults=True))
    # print(q_json)
    assert q_json == {"value": 1.0}


def test_pint():
    q1 = q.Length(value=1.0, unit=q.LengthUnit.milli_meter)
    # transform to pint
    q_pint = q1.to_pint()
    # transform back to QuantityValue
    q_ = q.QuantityValue.from_pint(q_pint)
    assert q1 == q_

    q2 = q.Length(value=1.0, unit=q.LengthUnit.meter)
    q3 = q1 + q2
    assert q3 == q.Length(value=1.001, unit=q.LengthUnit.meter)

    q31 = q1 * q2
    assert q31 == q.Area(value=1000.0, unit=q.AreaUnit.milli_meter_squared)

    q41 = q.Area(value=1.0, unit=q.AreaUnit.meter_squared)
    q42 = q.Area(value=1.0, unit=q.AreaUnit.milli_meter_squared)
    # 'square_meter' is not a valid unit for pint, but 'square_meter' is
    q43 = q41 + q42
    assert q43 == q.Area(value=1.000001, unit=q.AreaUnit.meter_squared)

    q5 = q.Length(value=2.0, unit=q.LengthUnit.meter)
    q25 = q2 / q5  # Dimensionless(value=0.5, unit=DimensionLessUnit.dimensionless)
    # will envoke QuantityValue.from_pint(), which will call unit_registry[unit_symbol]
    _ = q.VoltageRatio(value=q25.value)


def test_full_inventory_test():
    # test all QuantityValue classes
    warning_count = 0
    critical_warning_count = 0
    error_count = 0
    success_count = 0
    qu_reg = {}
    # build a list of all UnitEnums per QuantityValue class
    for ue, qv in quantity_registry.items():
        if qv not in qu_reg:
            qu_reg[qv] = []
        qu_reg[qv].append(ue)

    # build a list of all entr

    for qv in qu_reg.keys():
        # round-trip to and from pint
        # interate of the ue Enum members

        for ue in qu_reg[qv]:
            for u in ue:
                # create a QuantityValue object
                q1 = qv(value=1.0, unit=u)

                try:
                    # transform to pint
                    q_pint = q1.to_pint()
                    # transform back to QuantityValue
                    q_ = q.QuantityValue.from_pint(q_pint)
                    # assert q1 == q_
                    if q1 != q_:
                        # print(
                        # (
                        # f"Warning: {q1.value} {q1.unit} ",
                        # f"!= {q_.value} {q_.unit}",
                        # )
                        # )
                        warning_count += 1
                        # print(q1.to_pint())
                        # print(q_.to_pint())
                        if type(q1) is not type(q_):
                            print(
                                (
                                    f"Critical Warning: {q1.__class__.__name__} "
                                    f"and {q_.__class__.__name__} "
                                    f"have the same unit {u}"
                                )
                            )
                            critical_warning_count += 1
                            pass
                        elif q1.unit != q_.unit:
                            print(f"Warning: {q1.unit} was normalized to {q_.unit}")
                            pass
                        elif q1.value != q_.value:
                            print(f"Warning: Rounding error: {q1.value} vs {q_.value}")
                            pass
                        else:
                            print(f"Warning: Unknown error: {q1} vs {q_}")
                            pass

                    else:
                        success_count += 1
                except Exception as e:
                    # check if e is of type pint.errors.UndefinedUnitError
                    if isinstance(e, pint.errors.UndefinedUnitError):
                        print(f"Error: Missing unit {u} in pint")
                    elif isinstance(e, KeyError):
                        print(f"Error: Missing unit {u} in unit_registry")
                    else:
                        print(f"Error {e.__class__}: {q1} -> {q_pint} -> {q_}")
                        # print(re
                    error_count += 1
    print(
        (
            f"{success_count} successful, "
            f"{error_count} errors and {warning_count} warnings "
            f"({critical_warning_count} critical)"
        )
    )
    # 283 successful, 366 errors and 300 warnings # baseline
    # 295 successful, 292 errors and 362 warnings # fix inverse units
    # 346 successful, 122 errors and 481 warnings # fix SI prefixes
    # 359 successful, 105 errors and 485 warnings (444 critical) # fix all inverse


if __name__ == "__main__":
    test_addition()
    test_quantity_value()
    test_pint()
    test_full_inventory_test()

# pint_reg = pint.UnitRegistry()
# pint_q = pint_reg.Quantity(1.0, "petajoule")
# # not defined, try alias
# pint_q = pint_reg.Quantity(1.0, "joule 1 / kelvin 1 / kilogram squared 1 / pascal")
# print(pint_q)
# print(f"{pint_q:9f#Lx}")

# naming collisions?
# VaporPermeability
# NEONUnit
# CombinedNonEvaporativeHeatTransferCoefficient



import sys
from enum import Enum

if sys.version_info >= (3, 11):
    # Python 3.11 or higher
    from enum import EnumType
else:
    # Python < 3.11
    from enum import EnumMeta as EnumType

from typing import Any, Dict, List, Optional, Type, TypeVar, Union, overload

import pandas as pd

# pip install pint
import pint
import pint_pandas  # noqa: F401
from oold.model.v1 import LinkedBaseModelMetaClass as ModelMetaclass
from pydantic.v1 import Field, create_model
from pydantic.v1.fields import FieldInfo

from opensemantic import OswBaseModel

# import pint_pandas

# from osw.model.entity import Characteristic # pip install pint-pandas

QV = TypeVar("QV", bound="QuantityValue")


unit_registry: Dict[str, EnumType] = {}


class UnitEnumMetaclass(EnumType):
    def __new__(cls, clsname, bases, attrs):
        # print(attrs["__qualname__"], attrs)
        class_instance = EnumType.__new__(cls, clsname, bases, attrs)
        for key, value in attrs.items():
            if key not in ["__module__", "__qualname__", "_generate_next_value_"]:
                # register all enum values in the unit_registry
                if key in unit_registry:
                    pass
                    # warn(
                    #     (
                    #         f"Unit {key} already registered in "
                    #         "unit_registry, skip registration."
                    #     )
                    # )
                else:
                    unit_registry[key] = class_instance
        return class_instance


class UnitEnum(str, Enum, metaclass=UnitEnumMetaclass):
    pass


class Characteristic(OswBaseModel):
    """
    Elementary building block for object data schemas. Can inherit, reuse and/or define custom properties from other Characteristcs. Properties can have literal values or other complex values (objects) described by other Characteristics.
    """  # noqa: E501

    class Config:
        schema_extra = {
            "@context": [{"type": {"@id": "@type", "@type": "@id"}}],  # noqa: E501
            "uuid": "93ccae36-2435-42ce-ac6c-951450a81d47",  # noqa: E501
            "title": "Characteristic",  # noqa: E501
            "title*": {"en": "Characteristic", "de": "Charakteristik"},  # noqa: E501
            "description": "Elementary building block for object data schemas. Can inherit, reuse and/or define custom properties from other Characteristcs. Properties can have literal values or other complex values (objects) described by other Characteristics. ",  # noqa: E501
            "description*": {
                "en": "Elementary building block for object data schemas. Can inherit, reuse and/or define custom properties from other Characteristcs. Properties can have literal values or other complex values (objects) described by other Characteristics. ",  # noqa: E501
                "de": "Elementarer Baustein für Objekt-Datenschema. Kann Attribute von anderen Charakteristiken erben, übernehmen und/oder eigene definieren. Attribute können als Werte sowohl Literale als auch komplexe Objekte haben die von anderen Charakteristiken beschrieben werden.  ",  # noqa: E501
            },
        }

    type: Optional[List[str]] = Field(
        ["Category:OSW93ccae36243542ceac6c951450a81d47"],
        min_items=1,
        options={"hidden": True},
        propertyOrder=-1010,
        title="Types/Categories",
        title_={"de": "Typen/Kategorien"},
    )
    # should be optional:
    # uuid: UUID = Field(default_factory=uuid4, options={"hidden": True}, title="UUID")


ureg = pint.get_application_registry()

quantity_registry: Dict[EnumType, ModelMetaclass] = {}


class QuantityValueMetaclass(ModelMetaclass):
    def __new__(cls, clsname, bases, attrs):
        class_instance = super().__new__(cls, clsname, bases, attrs)
        # print(attrs["__qualname__"], attrs)
        if "unit" in attrs and attrs["unit"] is not None:
            # print(attrs["__annotations__"]["unit"].__dict__)#, attrs["unit"].__dict__)
            # print(attrs["unit"].__dict__["__objclass__"])
            # print(attrs["__annotations__"]["unit"].__args__[0])
            # register the mapping between the unit enum and the quantity class

            unit_field_type = None
            # check if FieldInfo was used for annotation
            if type(attrs["unit"]) is FieldInfo:
                _types = attrs["__annotations__"]["unit"].__args__
                # select the first type != None
                for _type in _types:
                    if _type is not None:
                        unit_field_type = _type
                        break
            else:
                unit_field_type = attrs["unit"].__dict__["__objclass__"]
            quantity_registry[unit_field_type] = class_instance
        return class_instance


class QuantityValue(Characteristic, metaclass=QuantityValueMetaclass):
    """
    A Quantity Value expresses the magnitude and kind of a quantity and is given by the product of a numerical value n and a unit of measure u (from qudt).
    """  # noqa: E501

    class Config:
        schema_extra = {
            "$comment": "Autogenerated section - do not edit. Generated from Category:Category Category:OSWffe74f291d354037b318c422591c5023",  # noqa: E501
            "uuid": "40829379-0663-4af9-92cf-9a1b18d772cf",
            "title": "QuantityValue",
            "title*": {"en": "Quantity Value"},
            "description": "A Quantity Value expresses the magnitude and kind of a quantity and is given by the product of a numerical value n and a unit of measure u (from qudt).",  # noqa: E501
            "description*": {
                "en": "A Quantity Value expresses the magnitude and kind of a quantity and is given by the product of a numerical value n and a unit of measure u (from qudt)."  # noqa: E501
            },
            "defaultProperties": ["type", "unit"],
            "@context": [
                "/wiki/Category:OSW93ccae36243542ceac6c951450a81d47?action=raw&slot=jsonschema",  # noqa: E501
                {"unit": "Property:HasUnit", "value": "Property:HasValue"},
            ],
            "$defs": {},
        }

    type: Optional[Any] = ["Category:OSW4082937906634af992cf9a1b18d772cf"]
    value: float = Field(
        ...,
        options={"grid_columns": 4},
        title="value",
        title_={"en": "Value", "de": "Wert"},
    )
    unit: Optional[UnitEnum] = Field(
        None,
        options={"grid_columns": 4},
        title="unit",
        title_={"en": "Unit", "de": "Einheit"},
    )

    # fmt: off
    @overload
    def __init__(self, value: float, unit: Optional[UnitEnum]) -> None:
        ...

    @overload
    def __init__(self, v: float, u: Optional[UnitEnum]) -> None:
        ...

    @overload
    def __init__(self, quantity_value: "QuantityValue") -> None:
        ...

    @overload
    def __init__(
        self, pint_quantity: pint.Quantity, quantity_type: Type["QuantityValue"]
    ) -> None:
        ...

    # @overload
    # def __init__(self, **data: Any) -> None:
    #     ...
    # fmt: on

    def __init__(self, *args, **kwargs) -> None:
        if "v" in kwargs or "u" in kwargs:
            # support for v, u as keyword arguments
            kwargs["value"] = kwargs.pop("v")
            kwargs["unit"] = kwargs.pop("u")
            # will raise error if alias and original names are mixed
        elif (
            len(args) == 2
            and isinstance(args[0], float)
            and isinstance(args[1], UnitEnum)
        ):
            # support for float, UnitEnum as positional arguments
            raise AttributeError(
                "QuantityValue.__init__() takes either a pint.Quantity and "
                "QuantityValue subclass as positional arguments, or a QuantityValue "
                "as positional arguments, but not a float and UnitEnum. These are "
                "keyword arguments only!"
            )
        elif (
            len(args) == 2
            and isinstance(args[0], pint.Quantity)
            and issubclass(args[1], QuantityValue)
        ):
            # support for pint.Quantity and Type[QuantityValue] as positional arguments
            qv_dict = QuantityValue.from_pint(
                quantity=args[0], simplify=True, quantity_type=args[1], return_dict=True
            )
            kwargs["value"] = qv_dict["value"]
            kwargs["unit"] = qv_dict["unit"]

        elif "pint_quantity" in kwargs and "quantity_type" in kwargs:
            # support for pint.Quantity and Type[QuantityValue] as keyword arguments
            qv_dict = QuantityValue.from_pint(
                quantity=kwargs.pop("pint_quantity"),
                simplify=True,
                quantity_type=kwargs.pop("quantity_type"),
                return_dict=True,
            )
            kwargs["value"] = qv_dict["value"]
            kwargs["unit"] = qv_dict["unit"]
        elif "pint_quantity" in kwargs and "quantity_type" not in kwargs:
            raise ValueError(
                "If 'pint_quantity' is provided, a 'quantity_type' to cast to must "
                "also be provided."
            )
        elif len(args) == 1 and len(kwargs) == 0 and isinstance([0], QuantityValue):
            # support for QuantityValue as positional argument
            kwargs = args[0].dict()
        elif "quantity_value" in kwargs:
            # support for QuantityValue as keyword argument
            quantity_value = kwargs.pop("quantity_value")
            kwargs["value"] = quantity_value.value
            kwargs["unit"] = quantity_value.unit

        super().__init__(**kwargs)

    @classmethod
    def get_pint_ureg_compatible_str(cls, string: str) -> str:
        pint_unit_name = string.replace("_", " ")
        # SI prefixes, see https://en.wikipedia.org/wiki/Metric_prefix
        prefixes = [
            "quetta",
            "ronna",
            "yotta",
            "zetta",
            "exa",
            "peta",
            "tera",
            "giga",
            "mega",
            "kilo",
            "hecto",
            "deca",
            "deci",
            "centi",
            "milli",
            "micro",
            "nano",
            "pico",
            "femto",
            "atto",
            "zepto",
            "yocto",
            "ronto",
            "quecto",
        ]

        for prefix in prefixes:
            pint_unit_name = pint_unit_name.replace(prefix + " ", prefix)
        if pint_unit_name.split(" ")[0] == "per":
            pint_unit_name = pint_unit_name.replace("per", "1 /")
        return pint_unit_name

    def to_pint(self) -> pint.Quantity:
        pint_unit_name = QuantityValue.get_pint_ureg_compatible_str(self.unit.name)

        return self.value * ureg[pint_unit_name]

    # fmt: on
    @classmethod
    @overload
    def from_pint(
        cls: Type[QV],
        quantity: pint.Quantity,
        simplify: bool = ...,
        quantity_type: Optional[Type[QV]] = ...,
        return_dict: bool = False,
    ) -> QV: ...

    @classmethod
    @overload
    def from_pint(
        cls: Type[QV],
        quantity: pint.Quantity,
        simplify: bool = ...,
        quantity_type: Optional[Type[QV]] = ...,
        return_dict: bool = True,
    ) -> Dict[str, Any]: ...

    # fmt: off

    @classmethod
    def from_pint(
        cls: Type[QV],
        quantity: pint.Quantity,
        simplify: bool = True,
        quantity_type: Optional[Type[QV]] = None,
        return_dict: bool = False,
    ) -> Union[QV, Dict[str, Any]]:
        # see also
        # https://pint.readthedocs.io/en/stable/getting/tutorial.html#simplifying-units
        # unit_symbol = "{:~P}".format(quantity.units)
        if quantity_type and not issubclass(quantity_type, QuantityValue):
            raise ValueError(
                f"Provided quantity_type '{quantity_type}' is not a subclass of "
                f"QuantityValue."
            )

        def original(quantity_: pint.Quantity):
            if simplify:
                value = (
                    f"{quantity_:9f#Lx}"  # 9f => round to 8 digits, '#' => simplify
                    # the unit
                )
            else:
                value = f"{quantity_:9fLx}"
            # e.g. \SI[]{1.0}{\kilo\gram\meter\per\ampere\squared\per\second\squared}
            # select the last curly brace
            unit_symbol = value.split("{")[-1].replace("}", "")
            # replace backslashes with underscores
            unit_symbol = unit_symbol.replace("\\", "_").strip("_")
            # nummeric_value = quantity.magnitude
            # simplifying the unit may change the scale
            nummeric_value = float(value.split("{")[1].split("}")[0])
            if len(unit_symbol) == 0:
                unit_symbol = "dimensionless"
            unit_class = unit_registry[unit_symbol]
            quantity_class = quantity_registry[unit_class]
            if quantity_type is not None:
                # if a specific quantity type is provided, use it instead of the
                # default one
                quantity_class = quantity_type
            if return_dict:
                return {
                    "value": nummeric_value,
                    "unit": unit_class[unit_symbol],
                    "quantity_type": quantity_class,
                }
            return quantity_class(value=nummeric_value, unit=unit_class[unit_symbol])

        def altered(quantity_: pint.Quantity):
            """
            Converts a Pint Quantity to base units (removing prefixes),
            and returns a LaTeX-formatted string with adjusted numeric value.
            """
            # Convert to base units to remove prefixes
            base_quantity = quantity_.to_base_units()

            # Format the numeric value and unit separately
            nummeric_value = base_quantity.magnitude
            unit_latex = f"{base_quantity.units:Lx}"
            if simplify:
                unit_latex = f"{base_quantity.units:#Lx}"

            unit_symbol_1 = unit_latex.split("{")[-1].replace("}", "")
            unit_symbol_2 = unit_symbol_1.replace("\\", "_").strip("_")
            if len(unit_symbol_2) == 0:
                unit_symbol_2 = "dimensionless"
            unit_class = unit_registry[unit_symbol_2]
            quantity_class = quantity_registry[unit_class]
            if quantity_type is not None:
                # if a specific quantity type is provided, use it instead of the
                # default one
                quantity_class = quantity_type
            if return_dict:
                return {
                    "value": nummeric_value,
                    "unit": unit_class[unit_symbol_2],
                    "quantity_type": quantity_class,
                }
            return quantity_class(value=nummeric_value, unit=unit_class[unit_symbol_2])

        try:
            return original(quantity)
        except KeyError:
            return altered(quantity)

    def to_base(self) -> "QuantityValue":
        """Converts the QuantityValue to its base unit."""
        pint_quantity = self.to_pint().to_base_units()
        return QuantityValue.from_pint(pint_quantity, simplify=False)

    def to_unit(self, unit: Union[UnitEnum, str]) -> "QuantityValue":
        """Converts the QuantityValue to the specified unit."""

        if isinstance(unit, UnitEnum):
            unit_str = QuantityValue.get_pint_ureg_compatible_str(unit.name)
        elif isinstance(unit, str):
            # Assuming the string is a valid unit from the pint unit registry
            unit_str = QuantityValue.get_pint_ureg_compatible_str(unit)
        else:
            raise ValueError(
                f"Invalid unit type: {type(unit)}. Must be UnitEnum or str.")

        pint_quantity = self.to_pint().to(unit_str)
        return QuantityValue.from_pint(
            quantity=pint_quantity, simplify=False, quantity_type=self.__class__)

    def __neg__(self) -> "QuantityValue":
        return QuantityValue.from_pint(-self.to_pint())

    def __pos__(self) -> "QuantityValue":
        return QuantityValue.from_pint(+self.to_pint())

    def __abs__(self) -> "QuantityValue":
        return QuantityValue.from_pint(abs(self.to_pint()))

    def __add__(self, other: "QuantityValue") -> "QuantityValue":
        res_pint = self.to_pint() + other.to_pint()
        return QuantityValue.from_pint(res_pint)

    def __sub__(self, other: "QuantityValue") -> "QuantityValue":
        res_pint = self.to_pint() - other.to_pint()
        return QuantityValue.from_pint(res_pint)

    # * operator
    def __mul__(self, other: Union["QuantityValue", float, int]) -> "QuantityValue":
        if not isinstance(other, (float, int)):
            other = other.to_pint()
        res_pint = self.to_pint() * other
        return QuantityValue.from_pint(res_pint)

    # / operator
    def __truediv__(self, other: Union["QuantityValue", float, int]) -> "QuantityValue":
        if not isinstance(other, (float, int)):
            other = other.to_pint()
        res_pint = self.to_pint() / other
        return QuantityValue.from_pint(res_pint)

    def __floordiv__(self, other: "QuantityValue") -> "QuantityValue":
        """Floor division is supported by pint only for dimensionless quantities and
        quantities with the same dimension"""
        if isinstance(other, QuantityValue):
            other = other.to_pint()
        return QuantityValue.from_pint(self.to_pint() // other)

    def __mod__(self, other: "QuantityValue") -> "QuantityValue":
        """Modulo is supported by pint only for dimensionless quantities and
        quantities with the same dimension"""
        if isinstance(other, QuantityValue):
            other = other.to_pint()
        return QuantityValue.from_pint(self.to_pint() % other)

    def __pow__(self, other: Union[int, float, "QuantityValue"]) -> "QuantityValue":
        # todo: test if supported with dimensionless unit / quantity
        if isinstance(other, QuantityValue):
            other = other.to_pint()
        return QuantityValue.from_pint(self.to_pint() ** other)

    def __eq__(self, other: "QuantityValue") -> bool:
        return self.to_pint() == other.to_pint()

    def __ne__(self, other: "QuantityValue") -> bool:
        return self.to_pint() != other.to_pint()

    def __ge__(self, other: "QuantityValue") -> bool:
        return self.to_pint() >= other.to_pint()

    def __gt__(self, other: "QuantityValue") -> bool:
        return self.to_pint() > other.to_pint()

    def __le__(self, other: "QuantityValue") -> bool:
        return self.to_pint() <= other.to_pint()

    def __lt__(self, other: "QuantityValue") -> bool:
        return self.to_pint() < other.to_pint()


QuantityValue.update_forward_refs()


class TabularData(OswBaseModel):
    # rows: List[Any] # convention for tabular data is a list of rows

    # consider https://stackoverflow.com/questions/51505504/pandas-nesting-dataframes # noqa: E501
    # consider https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.attrs.html # noqa: E501

    def to_df(self) -> pd.DataFrame:
        series = []
        row_class = self.__class__.__fields__["rows"].type_
        for attr in row_class.__fields__.keys():
            q_name = (
                row_class.__fields__[attr]
                .type_.__fields__["unit"]
                .default.name.replace("_", " ")
            )
            # q_pint = ureg[q_name]
            s = pd.Series(
                [
                    (
                        getattr(m, attr).to_pint().to(q_name).magnitude
                        if getattr(m, attr, None) is not None
                        else None
                    )
                    for m in self.rows
                ],
                dtype="pint[" + q_name + "]",
                name=attr,
            )
            series.append(s)
        rows = {s.name: s for s in series}
        return pd.DataFrame(rows)

    @classmethod
    def from_df(cls, df: pd.DataFrame):
        rows = []
        row_class = OswBaseModel
        if "rows" in cls.__fields__:
            row_class = cls.__fields__["rows"].type_

        additional_fields = {}
        for key in df.columns:
            # print(df[key].pint.__dict__)
            if key not in row_class.__fields__.keys():
                # raise ValueError(f"Column '{key}' not found in '{row_class.__name__}'") # noqa: E501
                quantity = QuantityValue.from_pint(
                    1 * getattr(df.dtypes, key).units
                ).__class__
                additional_fields[key] = (quantity, ...)

        if len(additional_fields) > 0:
            row_class = create_model(
                row_class.__name__ + "Extended", **additional_fields, __base__=row_class
            )
            cls = create_model(
                cls.__name__ + "Extended",
                **{"rows": (List[row_class], ...)},
                __base__=cls,
            )

        # convert all columns to the default unit
        for attr in row_class.__fields__.keys():
            q_name = (
                row_class.__fields__[attr]
                .type_.__fields__["unit"]
                .default.name.replace("_", " ")
            )
            q_pint = ureg[q_name]
            df[attr] = df[attr].pint.to(q_pint)

        for i, row in df.iterrows():
            # create a dictionary with the values of the row
            # using the column names as keys
            d = {key: {"value": row[key].magnitude} for key in df.columns}
            m = row_class(**d)
            rows.append(m)
        return cls(rows=rows)



"""This is an integration test for the QuantityValue functionality"""

import json
from typing import List, Optional

from scipy.stats import linregress

from opensemantic import OswBaseModel
from opensemantic.characteristics.quantitative import (
    Area,
    AreaUnit,
    Diameter,
    DimensionlessUnit,
    Force,
    ForcePerAreaUnit,
    ForceUnit,
    Length,
    LengthUnit,
    LinearStrain,
    ModulusOfElasticity,
    QuantityValue,
    Stress,
    TabularData,
    Thickness,
    Width,
)

# Do we have to adapt VSCode settings to include the package index?
# "python.analysis.packageIndexDepths": [
#       {"name": "opensemantic.characteristics.quantitative",
#       "depth": 4, "includeAllSymbols": true}
# ]


def test_pint():

    q = Length(value=1.0, unit=LengthUnit.milli_meter)
    # transform to pint
    q_pint = q.to_pint()
    # transform back to QuantityValue
    q_ = QuantityValue.from_pint(q_pint)
    assert q == q_

    q2 = Length(value=1.0, unit=LengthUnit.meter)
    q3 = q + q2
    assert q3 == Length(value=1.001, unit=LengthUnit.meter)

    q31 = q * q2
    assert q31 == Area(value=1000.0, unit=AreaUnit.milli_meter_squared)

    q41 = Area(value=1.0, unit=AreaUnit.meter_squared)
    q42 = Area(value=1.0, unit=AreaUnit.milli_meter_squared)
    # 'square_meter' is not a valid unit for pint, but 'square_meter' is
    q43 = q41 + q42
    assert q43 == Area(value=1.000001, unit=AreaUnit.meter_squared)


def test_quantityvalue_magic_methods():

    l1 = Length(value=10, unit=LengthUnit.meter)
    l2 = Length(value=3, unit=LengthUnit.meter)
    # __neg__
    neg = -l1
    assert isinstance(neg, Length)
    assert neg.value == -10
    # __pos__
    pos = +l1
    assert isinstance(pos, Length)
    assert pos.value == 10
    # __abs__
    absval = abs(Length(value=-5, unit=LengthUnit.meter))
    assert isinstance(absval, Length)
    assert absval.value == 5
    # __add__
    add = l1 + l2
    assert isinstance(add, Length)
    assert add.value == 13
    # __sub__
    sub = l1 - l2
    assert isinstance(sub, Length)
    assert sub.value == 7
    # __mul__
    mul = l1 * l2
    assert isinstance(mul, Area)
    # __truediv__
    truediv = l1 / l2
    assert isinstance(truediv, QuantityValue)
    assert truediv.unit == DimensionlessUnit.dimensionless
    # __floordiv__
    floordiv = l1 // l2
    assert isinstance(floordiv, QuantityValue)
    assert floordiv.value == 3.0
    # __mod__
    mod = l1 % l2
    assert isinstance(mod, QuantityValue)
    assert mod.value == 1.0
    # __pow__
    pow_result = l1**2
    assert isinstance(pow_result, Area)
    # __eq__
    eq = l1 == Length(value=10, unit=LengthUnit.meter)
    assert eq is True
    # __ne__
    ne = l1 != l2
    assert ne is True
    # __ge__
    ge = l1 >= l2
    assert ge is True
    # __gt__
    gt = l1 > l2
    assert gt is True
    # __le__
    le = l1 <= Length(value=10, unit=LengthUnit.meter)
    assert le is True
    # __lt__
    lt = l1 < Length(value=11, unit=LengthUnit.meter)
    assert lt is True


def test_export():

    q = Length(value=1.0, unit=LengthUnit.milli_meter)

    q_json = json.loads(q.json(exclude_none=True))
    print(q_json)
    assert q_json == {
        "type": ["Category:OSWee9c7e5c343e542cb5a8b4648315902f"],
        "value": 1.0,
        "unit": str(
            "Item:OSWf101d25e944856e3bd4b4c9863db7de2"
            "#OSW322dec469be75aedb008b3ebff29db86"
        ),
    }

    q = Length(value=1.0, unit=LengthUnit.meter)

    q_json = json.loads(q.json(exclude_none=True))
    print(q_json)
    assert q_json == {
        "type": ["Category:OSWee9c7e5c343e542cb5a8b4648315902f"],
        "value": 1.0,
        "unit": "Item:OSWf101d25e944856e3bd4b4c9863db7de2",
    }

    q_json = json.loads(q.json(exclude_none=True, exclude_defaults=True))
    print(q_json)
    assert q_json == {"value": 1.0}

    ln = Length(value=0.1)
    w = Width(value=200, unit=LengthUnit.milli_meter)
    a = ln * w
    print(a)

    json_dict = a.dict()
    print(json_dict)
    assert json_dict["type"] == ["Category:OSW1fcf1694712e5684885071efdf775bd9"]
    assert json_dict["value"] == 20000.0
    assert json_dict["unit"] == (
        "Item:OSWd10e5841c68e5aad94b481b58ef9dfb9"
        "#OSWeca22bf4270853038ef3395bd6dd797b"
    )

    # _a = QuantityValue(**json_dict)

    _a = a.to_base()
    json_dict = _a.dict(exclude_none=True, exclude_defaults=True)
    print(json_dict)
    assert json_dict["value"] == 0.02
    assert len(json_dict.keys()) == 1

    __a = Area(**json_dict)
    assert __a == _a

    # not supported yet
    # jsonld_dict = a.to_jsonld()
    # print(json.dumps(jsonld_dict, indent=2))

    # a2 = QuantityValue.from_jsonld(jsonld_dict)
    # print(a2)

    a3 = Area(value=4.0, unit=AreaUnit.meter_squared)
    a4 = a3.to_unit(AreaUnit.centi_meter_squared)
    assert a4.unit == AreaUnit.centi_meter_squared
    assert a4.value == 100**2 * a3.value
    assert a3 == a4

    json_dict4 = a4.dict(exclude_none=True, exclude_defaults=True)
    print(json_dict4)
    assert json_dict4["value"] == 40000
    assert json_dict4["unit"] == (
        "Item:OSWd10e5841c68e5aad94b481b58ef9dfb9#OSWe36916dd7a34557b8a52c38d6dd7b832"
    )
    assert len(json_dict4.keys()) == 2


def test_pandas():

    class Measurement(OswBaseModel):
        length: Length
        width: Width
        area: Optional[Area] = None

    class MeasurementData(TabularData):
        rows: List[Measurement]

    measurement_data = MeasurementData(
        rows=[
            Measurement(length=Length(value=1.0), width=Width(value=2.0)),
            Measurement(length=Length(value=3.0), width=Width(value=4.0)),
        ]
    )
    df = measurement_data.to_df()
    print(df)
    df["area"] = df["length"] * df["width"]
    print(df)
    measurement_data2 = MeasurementData.from_df(df)
    print(measurement_data2)
    measurement_data3 = TabularData.from_df(df)
    print(measurement_data3.json(exclude_none=True, exclude_defaults=True, indent=2))
    # print(measurement_data3.__class__.schema_json(indent=2))


def test_tensile_test():

    class TensileTestSpecimen(OswBaseModel):
        length: Length
        width: Width
        thickness: Length
        cross_section_area: Optional[Area] = None
        e_mod: Optional[ModulusOfElasticity] = None

    class TensileTestResultRow(OswBaseModel):
        elongation: Length
        force: Force
        strain: Optional[LinearStrain] = None
        stress: Optional[Stress] = None

    class TensileTestResult(TabularData):
        rows: List[TensileTestResultRow]
        linear_region: Optional[LinearStrain] = LinearStrain(value=0.002)

    class TensileTestDataset(OswBaseModel):
        specimen: TensileTestSpecimen
        result: TensileTestResult

    def tensile_test_analysis(dataset: TensileTestDataset):
        # Calculate cross section area if not provided
        if dataset.specimen.cross_section_area is None:
            dataset.specimen.cross_section_area = (
                dataset.specimen.width * dataset.specimen.thickness
            )

        # Calculate stress and strain - slow iteration
        # for row in dataset.result.rows:
        #     #row.stress = Stress(value=row.force.value / dataset.specimen.cross_section_area.value, unit=row.force.unit)   # noqa: E501
        #     #row.strain = LinearStrain(value=row.force.value / (dataset.specimen.e_mod.value * dataset.specimen.cross_section_area.value), unit=row.force.unit)   # noqa: E501
        #     row.stress = row.force / dataset.specimen.cross_section_area   # noqa: E501

        # Calculate stress and strain - fast as pandas DataFrame operation
        df = dataset.result.to_df()
        df["strain"] = df["elongation"] / dataset.specimen.length.to_pint()
        df["stress"] = df["force"] / dataset.specimen.cross_section_area.to_pint()

        # make a linear fit on the linear region of the stress-strain curve
        # to find the modulus of elasticity
        linear_region = df[df["strain"] <= dataset.result.linear_region.to_pint()]

        slope, intercept, r_value, p_value, std_err = linregress(
            linear_region["strain"].pint.to_base_units().pint.magnitude,
            linear_region["stress"].pint.to_base_units().pint.magnitude,
        )

        slope = (
            slope
            * linear_region["stress"].pint.to_base_units().pint.units
            / linear_region["strain"].pint.to_base_units().pint.units
        )  # noqa: E501
        dataset.specimen.e_mod = ModulusOfElasticity.from_pint(slope.to("Pa"))

        dataset.result = TensileTestResult.from_df(df)

        return dataset

    # Example usage
    specimen = TensileTestSpecimen(
        length=Length(value=100, unit=LengthUnit.milli_meter),
        width=Width(value=10, unit=LengthUnit.milli_meter),
        thickness=Thickness(value=10, unit=LengthUnit.milli_meter),
    )
    result = TensileTestResult(
        rows=[
            TensileTestResultRow(
                force=Force(value=1.0000, unit=ForceUnit.kilo_newton),
                elongation=Length(value=0.10, unit=LengthUnit.milli_meter),
            ),  # noqa: E501
            TensileTestResultRow(
                force=Force(value=1.5050, unit=ForceUnit.kilo_newton),
                elongation=Length(value=0.15, unit=LengthUnit.milli_meter),
            ),  # noqa: E501
            TensileTestResultRow(
                force=Force(value=2.0000, unit=ForceUnit.kilo_newton),
                elongation=Length(value=0.20, unit=LengthUnit.milli_meter),
            ),  # noqa: E501
        ]
    )
    dataset = TensileTestDataset(specimen=specimen, result=result)
    analyzed_dataset = tensile_test_analysis(dataset)
    # print(analyzed_dataset.json(exclude_none=True, indent=2))
    assert analyzed_dataset.specimen.e_mod == ModulusOfElasticity(
        value=10.0, unit=ForcePerAreaUnit.giga_pascal
    )


def test_init():
    # Overload 1: __init__(self, value: float, unit: Optional[UnitEnum])
    l1 = Length(value=5.0, unit=LengthUnit.meter)
    assert isinstance(l1, Length)
    assert l1.value == 5.0
    assert l1.unit == LengthUnit.meter

    # Overload 2: __init__(self, v: float, u: Optional[UnitEnum])
    l2 = Length(v=10.0, u=LengthUnit.milli_meter)
    assert isinstance(l2, Length)
    assert l2.value == 10.0
    assert l2.unit == LengthUnit.milli_meter

    # Overload 3: __init__(self, quantity_value: "QuantityValue")
    d3 = Diameter(quantity_value=l1)
    assert isinstance(d3, Diameter)
    assert d3.value == l1.value
    assert d3.unit == l1.unit

    # Overload 4: __init__(self, pint_quantity: pint.Quantity,
    #                      quantity_type: Type[QuantityValue])
    import pint

    ureg = pint.get_application_registry()
    pq = 2.5 * ureg.meter
    l4 = Length(pint_quantity=pq, quantity_type=Length)
    assert isinstance(l4, Length)
    assert l4.value == 2.5
    assert l4.unit == LengthUnit.meter

    # Overload 5: __init__(self, **data: Any)
    l5 = Length(**{"value": 7.0, "unit": LengthUnit.meter})
    assert isinstance(l5, Length)
    assert l5.value == 7.0
    assert l5.unit == LengthUnit.meter


def test_to_unit():
    # Test conversion from meters to millimeters
    length = Length(value=1.0, unit=LengthUnit.meter)
    l_mm = length.to_unit(LengthUnit.milli_meter)
    assert isinstance(l_mm, Length)
    assert l_mm.unit == LengthUnit.milli_meter
    assert l_mm.value == 1000.0

    # Test conversion using string
    l_cm = length.to_unit("centi_meter")
    assert isinstance(l_cm, Length)
    assert l_cm.unit.name == "centi_meter"
    assert l_cm.value == 100.0


if __name__ == "__main__":
    test_pint()
    test_export()
    test_pandas()
    test_tensile_test()
    test_quantityvalue_magic_methods()
    test_init()



from importlib.metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = "opensemantic.characteristics.quantitative"
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError

from opensemantic.characteristics.quantitative._model import *  # noqa
from opensemantic.characteristics.quantitative._static import (  # noqa
    QuantityValue,
    TabularData,
)
